/*
 * Provides an average of values.
 *
 * Author: J. Lewis Muir
 */

#include <stdio.h>
#include <stdlib.h>

#include <epicsAssert.h>
#include <epicsStdio.h>
#include <epicsString.h>
#include <epicsExport.h>
#include <iocsh.h>

#include <asynPortDriver.h>
#include "AmyAverage.h"

AmyAverage::AmyAverage(const char *portName, int numAddrs, int maxNumValues,
      int integerStorageAndArithmetic)
    : asynPortDriver(portName, numAddrs < 1 ? 1 : numAddrs,
        NUM_AMY_AVERAGE_PARAMS,
        asynInt32Mask | asynFloat64Mask | asynDrvUserMask,
        asynInt32Mask | asynFloat64Mask,
        ASYN_MULTIDEVICE, 1, 0, 0) {
  int i;
  int validMaxNumValues;
  int status = asynSuccess;
  const char *funcName = "AmyAverage::AmyAverage";

  validMaxNumValues = maxNumValues < 1 ? 1 : maxNumValues;
  this->integerStorageAndArithmetic = integerStorageAndArithmetic;

  windowIndex = NULL;
  windowFull = NULL;
  integerValues = NULL;
  floatValues = NULL;

  windowIndex = (int *)calloc(maxAddr, sizeof(*windowIndex));
  if (windowIndex == NULL) {
    fprintf(stderr, "%s: failed to allocate memory for windowIndex array\n",
      funcName);
    return;
  }
  windowFull = (int *)calloc(maxAddr, sizeof(*windowFull));
  if (windowFull == NULL) {
    fprintf(stderr, "%s: failed to allocate memory for windowFull array\n",
      funcName);
    return;
  }
  for (i = 0; i < maxAddr; i++) {
    windowIndex[i] = 0;
    windowFull[i] = 0;
  }

  if (this->integerStorageAndArithmetic) {
    integerValues = (epicsInt32 **)calloc(maxAddr, sizeof(*integerValues));
    if (integerValues == NULL) {
      fprintf(stderr, "%s: failed to allocate memory for array of "
        "pointers to integer value arrays\n", funcName);
      return;
    }
    for (i = 0; i < maxAddr; i++) {
      integerValues[i] = (epicsInt32 *)calloc(validMaxNumValues,
        sizeof(**integerValues));
      if (integerValues[i] == NULL) {
        fprintf(stderr, "%s: failed to allocate memory for addr=%d "
          "integer values\n", funcName, i);
        return;
      }
    }
  } else {
    floatValues = (epicsFloat64 **)calloc(maxAddr, sizeof(*floatValues));
    if (floatValues == NULL) {
      fprintf(stderr, "%s: failed to allocate memory for array of "
        "pointers to floating-point value arrays\n", funcName);
      return;
    }
    for (i = 0; i < maxAddr; i++) {
      floatValues[i] = (epicsFloat64 *)calloc(validMaxNumValues,
        sizeof(**floatValues));
      if (floatValues[i] == NULL) {
        fprintf(stderr, "%s: failed to allocate memory for addr=%d "
          "floating-point values\n", funcName, i);
        return;
      }
    }
  }

  for (i = 0; i < maxAddr; i++) {
    status |= createParam(i, P_ValueIntString,     asynParamInt32,
      &P_ValueInt);
    status |= createParam(i, P_ValueFltString,     asynParamFloat64,
      &P_ValueFlt);
    status |= createParam(i, P_ArithMeanIntString, asynParamInt32,
      &P_ArithMeanInt);
    status |= createParam(i, P_ArithMeanFltString, asynParamFloat64,
      &P_ArithMeanFlt);
    status |= createParam(i, P_MovingString,       asynParamInt32,
      &P_Moving);
    status |= createParam(i, P_ResetString,        asynParamInt32,
      &P_Reset);
    status |= createParam(i, P_WindowSizeString,   asynParamInt32,
      &P_WindowSize);
    status |= createParam(i, P_MaxNumValuesString, asynParamInt32,
      &P_MaxNumValues);
  }

  for (i = 0; i < maxAddr; i++) {
    status |= setIntegerParam(i, P_ValueInt,     0);
    status |= setDoubleParam( i, P_ValueFlt,     0.0);
    status |= setIntegerParam(i, P_ArithMeanInt, 0);
    status |= setDoubleParam( i, P_ArithMeanFlt, 0.0);
    status |= setIntegerParam(i, P_Moving,       0);
    status |= setIntegerParam(i, P_Reset,        0);
    status |= setIntegerParam(i, P_WindowSize,   validMaxNumValues);
    status |= setIntegerParam(i, P_MaxNumValues, validMaxNumValues);
  }

  if (status) {
    fprintf(stderr, "%s: failed to initialize parameter values\n", funcName);
    return;
  }
}

AmyAverage::~AmyAverage() {
  int i;

  if (windowIndex != NULL) free(windowIndex);
  if (windowFull != NULL) free(windowFull);
  if (integerValues != NULL) {
    for (i = 0; i < maxAddr; i++)
      if (integerValues[i] != NULL) free(integerValues[i]);
    free(integerValues);
  }
  if (floatValues != NULL) {
    for (i = 0; i < maxAddr; i++)
      if (floatValues[i] != NULL) free(floatValues[i]);
    free(floatValues);
  }
}

asynStatus AmyAverage::writeInt32(asynUser *pasynUser, epicsInt32 value) {
  int function = pasynUser->reason;
  int addr = 0;
  int status = asynSuccess;
  const char *paramName = "UNKNOWN";
  const char *funcName = "AmyAverage::writeInt32";

  status = getAddress(pasynUser, &addr);
  if (status != asynSuccess) return (asynStatus)status;

  if (function == P_ValueInt) setValue(addr, value, (epicsFloat64)value);
  else if (function == P_Moving) setMoving(addr, value);
  else if (function == P_Reset) reset(addr);
  else if (function == P_WindowSize) setWindowSize(addr, value);

  status |= callParamCallbacks(addr, addr);

  getParamName(function, &paramName);
  if (status)
    epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
      "%s: status=%d, function=%d, name=%s, value=%d, addr=%d", funcName,
      status, function, paramName, value, addr);
  else
    asynPrint(pasynUser, ASYN_TRACEIO_DRIVER,
      "%s: function=%d, name=%s, value=%d, addr=%d\n", funcName, function,
      paramName, value, addr);

  return (asynStatus)status;
}

asynStatus AmyAverage::writeFloat64(asynUser *pasynUser, epicsFloat64 value) {
  int function = pasynUser->reason;
  int addr = 0;
  int status = asynSuccess;
  const char *paramName = "UNKNOWN";
  const char *funcName = "AmyAverage::writeFloat64";

  status = getAddress(pasynUser, &addr);
  if (status != asynSuccess) return (asynStatus)status;

  if (function == P_ValueFlt)
      setValue(addr, (epicsInt32)value, integerStorageAndArithmetic ?
        (epicsFloat64)(epicsInt32)value : value);

  status |= callParamCallbacks(addr, addr);

  getParamName(function, &paramName);
  if (status)
    epicsSnprintf(pasynUser->errorMessage, pasynUser->errorMessageSize,
      "%s: status=%d, function=%d, name=%s, value=%f, addr=%d", funcName,
      status, function, paramName, value, addr);
  else
    asynPrint(pasynUser, ASYN_TRACEIO_DRIVER,
      "%s: function=%d, name=%s, value=%f, addr=%d\n", funcName, function,
      paramName, value, addr);

  return (asynStatus)status;
}

void AmyAverage::setValue(int addr, epicsInt32 integerValue,
    epicsFloat64 floatValue) {
  int status;

  status  = setIntegerParam(addr, P_ValueInt, integerValue);
  status |= setDoubleParam(addr, P_ValueFlt, floatValue);
  assert(status == asynSuccess);
  if (integerStorageAndArithmetic)
    integerValues[addr][windowIndex[addr]] = integerValue;
  else
    floatValues[addr][windowIndex[addr]] = floatValue;
  windowIndex[addr] = (windowIndex[addr] + 1) % getWindowSize(addr);
  if (windowIndex[addr] == 0) windowFull[addr] = 1;
  computeAndSetArithMeanIfNeeded(addr);
}

void AmyAverage::computeAndSetArithMeanIfNeeded(int addr) {
  int status;
  int regionSize;
  epicsInt32 integerArithMean;
  epicsFloat64 floatArithMean;

  if (!isMoving(addr) && !windowFull[addr]) return;
  regionSize = windowFull[addr] ? getWindowSize(addr) : windowIndex[addr];
  if (integerStorageAndArithmetic) {
    integerArithMean = computeIntegerArithMean(addr, regionSize);
    floatArithMean = (epicsFloat64)integerArithMean;
  } else {
    floatArithMean = computeFloatArithMean(addr, regionSize);
    integerArithMean = (epicsInt32)floatArithMean;
  }
  status  = setIntegerParam(addr, P_ArithMeanInt, integerArithMean);
  status |= setDoubleParam(addr, P_ArithMeanFlt, floatArithMean);
  assert(status == asynSuccess);
  if (!isMoving(addr)) windowFull[addr] = 0;
}

epicsInt32 AmyAverage::computeIntegerArithMean(int addr, int regionSize) {
  int i;
  epicsInt32 sum = 0;

  for (i = 0; i < regionSize; i++) sum += integerValues[addr][i];
  return sum / (epicsInt32)regionSize;
}

epicsFloat64 AmyAverage::computeFloatArithMean(int addr, int regionSize) {
  int i;
  epicsFloat64 sum = 0;

  for (i = 0; i < regionSize; i++) sum += floatValues[addr][i];
  return sum / (epicsFloat64)regionSize;
}

int AmyAverage::isMoving(int addr) {
  asynStatus status;
  int moving = 0;

  status = getIntegerParam(addr, P_Moving, &moving);
  assert(status == asynSuccess);
  return moving;
}

void AmyAverage::setMoving(int addr, int moving) {
  asynStatus status;
  int validMoving;

  validMoving = moving == 0 ? 0 : 1;
  if (validMoving == isMoving(addr)) return;
  status = setIntegerParam(addr, P_Moving, validMoving);
  assert(status == asynSuccess);
  reset(addr);
}

void AmyAverage::reset(int addr) {
  int status;
  epicsInt32 integerValue;
  epicsFloat64 floatValue;

  status  = getIntegerParam(addr, P_ValueInt, &integerValue);
  status |= getDoubleParam(addr, P_ValueFlt, &floatValue);
  assert(status == asynSuccess);
  windowIndex[addr] = 0;
  windowFull[addr] = 0;
}

epicsInt32 AmyAverage::getWindowSize(int addr) {
  asynStatus status;
  epicsInt32 size;

  status = getIntegerParam(addr, P_WindowSize, &size);
  assert(status == asynSuccess);
  return size;
}

void AmyAverage::setWindowSize(int addr, epicsInt32 size) {
  asynStatus status;
  epicsInt32 maxNumValues;
  epicsInt32 validSize;

  if (size == getWindowSize(addr)) return;
  maxNumValues = getMaxNumValues(addr);
  validSize = size < 1 ? 1 : (size > maxNumValues ? maxNumValues : size);
  status = setIntegerParam(addr, P_WindowSize, validSize);
  assert(status == asynSuccess);
  reset(addr);
}

epicsInt32 AmyAverage::getMaxNumValues(int addr) {
  asynStatus status;
  epicsInt32 maxNumValues;

  status = getIntegerParam(addr, P_MaxNumValues, &maxNumValues);
  assert(status == asynSuccess);
  return maxNumValues;
}

/*
 * C function for configuring the AmyAverage asyn port driver.
 *
 * portName                     Name of asyn port to be used by this driver
 * numAddrs                     Number of addresses (i.e. virtual devices)
 * maxNumValues                 Maximum number of values that may be averaged
 *                              (>= 1)
 * integerStorageAndArithmetic  Whether to store values to be averaged as
 *                              integers and perform integer arithmetic or to
 *                              use floating-point storage and arithmetic
 *                              (0=floating-point, 1=integer)
 */
extern "C" int amyAverageConfig(const char *portName, int numAddrs,
    int maxNumValues, int integerStorageAndArithmetic) {
  new AmyAverage(portName, numAddrs, maxNumValues,
    integerStorageAndArithmetic);
  return 0;
}


/*
 * AmyAverage asyn port driver EPICS iocsh interface.
 */
extern "C" {
static const iocshArg amyAverageConfigArg0  = {
  "Port name", iocshArgString};
static const iocshArg amyAverageConfigArg1  = {
  "Number of addresses (i.e. virtual devices)", iocshArgInt};
static const iocshArg amyAverageConfigArg2  = {
  "Max number of values (>= 1)", iocshArgInt};
static const iocshArg amyAverageConfigArg3  = {
  "Store and compute values as integers (0=no, 1=yes)", iocshArgInt};
static const iocshArg *const amyAverageConfigArgs[4] = {
  &amyAverageConfigArg0,
  &amyAverageConfigArg1,
  &amyAverageConfigArg2,
  &amyAverageConfigArg3};
static const iocshFuncDef configAmyAverage = {
  "amyAverageConfig", 4, amyAverageConfigArgs};
static void configAmyAverageCallFunc(const iocshArgBuf *args) {
  amyAverageConfig(args[0].sval, args[1].ival, args[2].ival, args[3].ival);
}
static void amyRegister(void) {
  iocshRegister(&configAmyAverage, configAmyAverageCallFunc);
}
epicsExportRegistrar(amyRegister);
}
